require(wmtsa)

###############################################################################
## Utilitaires :

#plotting routine :
plotThis = function(data,cl=rep(1,nrow(data)),range=c(min(data),max(data)), ...) {
	for (i in 1:nrow(data)) {
		plot(data[i,],col=cl[i],ylim=range,type="l", ...)
		par(new=TRUE)
	}
}

#increase the number of discretisation points if needed :
expandSplines = function(data,nbEvals) {
	N = nrow(data)
	expData = matrix(nrow=N,ncol=nbEvals)
	for (i in 1:N) {
		expData[i,] = (spline(x=seq(to=ncol(data)),y=data[i,],n=nbEvals))$y
	}
	return (expData)
}

#distortion in reference to a data matrix, and some clusters' centers
distortion = function(dataToDistor,centers,plotting=FALSE) {
	nclusts = nrow(centers)
	allDists = rep(0.0,nclusts)
	distor = 0.0
	n = nrow(dataToDistor)
	cl = rep(0,n)
	seqVect = seq(to=nclusts)
	for (i in 1:n) {
		for (j in 1:nclusts)
			allDists[j] = sum((dataToDistor[i,]-centers[j,])^2)
		distor = distor + min(allDists) / ncol(dataToDistor)
		if (plotting) cl[i] = as.integer(seqVect[rank(allDists)==1])
	}
	rg = c(min(dataToDistor),max(dataToDistor))
	if (plotting) plotThis(dataToDistor,cl,rg)
#	return (sqrt(distor / n)) #mon ancienne version "normalisee"
	return (distor / n)
}


###############################################################################
## Basis generation :

# select the sub-basis of most variable coefficients
basisMostVar = function(basis,dataForRef,nbCoefs,verbose) {
	F = dataForRef %*% t(basis)
	rnCoefs = ncol(F)
	if (verbose && rnCoefs < nbCoefs) print(paste("    real number of coefficients: ",rnCoefs))

	if (nbCoefs >= rnCoefs) return (basis)
	V = rep(0,rnCoefs)
	for (i in 1:rnCoefs) V[i] = var(F[,i])
	seqVect=seq(to=rnCoefs)
	indices = seqVect[rank(V)>(rnCoefs-nbCoefs)]
	
	result = matrix(nrow=length(indices),ncol=ncol(dataForRef))
	counter = 1
	for (i in indices) {
		result[counter,] = basis[i,]
		counter = counter + 1
	}
	return (result)
}

#return distortion, centers and clusters composition for a given basis and signal matrix :
allParamsBasis = function(data, basis, nclusts, nbCoefs, dataToDistor) {
	F = data %*% t(basis) #reconst = F %*% basis
	k = kmeans(F,nclusts,nstart=100,iter.max=500)

	#prepare the output with all information available
	centers = k$centers %*% basis
	distor = distortion(dataToDistor,centers)
	result = list('clusts'=k$cluster,'ctrs'=centers,'disto'=distor)
	return (result)
}

#generate Fourier basis on an interval (card == nbCoefs)
genFourier = function(dataForRef,nbCoefs,withVar=TRUE) {
	D = ncol(dataForRef)
	res = matrix(nrow=floor(D/2)+1,ncol=D)
	res[1,] = rep(1.0/(sqrt(D)),D)
	times = seq(to=D)
	for (i in 1:floor(D/4)) {
		res[2*i,] = cos( (2*pi*i*times) / (D-1) )
		res[2*i+1,] = sin( (2*pi*i*times) / (D-1) )
	}
	# orthonormalize the basis :
	for (i in 2:nrow(res))
		res[i,] = res[i,] / sqrt(sum(res[i,]^2))
	if (!withVar) return (res[1:min(nbCoefs,nrow(res)),])

	#get the max variance coefficients according to data :
	return (basisMostVar(res,dataForRef,nbCoefs,verbose=FALSE))
}

########## Basic wavelets :

# get a basis from a wavelet packet transform and a basis tree
crystalsToBasis = function(W, lvl, os, filt) {
	n = length(lvl) # number of packet levels
	D = length(W$data$w0.0) # length of original signal
	bbasis = matrix(nrow=n, ncol=D)
	wm = wavMRD(wavDWPT(W, wavelet = filt),level=lvl, osc=os)
	for (i in 1:n)
		bbasis[i,] = wm[[i]]
	return (bbasis)
}

# "normal" wavelet basis expansion
simpleWavBasis = function(dataForRef, lvl, filt) {
	N = nrow(dataForRef) ; nd = rep(0,N)
	for (i in 1:N) nd[i] = sum(dataForRef[i,]^2)
	R = rank(nd) ; seqVect = seq(to=N)
	med = dataForRef[seqVect[R==N%/%2],]
	W <- wavDWPT(med, n.level=lvl, wavelet = filt)

	D = length(W$data$w0.0) # length of original signal
	wm = wavMRD(wavDWPT(W, wavelet = filt))
	nbFuncs = ncol(wm)
	basis = matrix(nrow=nbFuncs, ncol=D)

	for (i in 1:nbFuncs)
		basis[i,] = wm[[i]]

	# orthonormalize the basis :
	for (i in 1:nbFuncs)
		basis[i,] = basis[i,] / sqrt(sum(basis[i,]^2))
	return (basis)
}

#seuillage manuel, comme pour K-L et autres :
threshWav = function(dataForRef, lvl, filt, nbCoefs,withVar=TRUE) {
	s = simpleWavBasis(dataForRef,lvl,filt)
	if (!withVar) {
		return (s[1:min(nrow(s),nbCoefs),])
	}
	return (basisMostVar(s,dataForRef,nbCoefs,verbose=FALSE))
}

########## "Best" wavelets :

# define an entropy cost functional
"entropy" <- function(x) {
	iz <- which(x==0)
	z <- -x^2 * log(x^2)
	if (length(iz))
		z[iz] <- 0
	return (sum(z))
}

# define a distance-based cost (marche ?! ..)
"distanc" <- function(x) {
	d = dist(x,method='euclidian')
	q = quantile(x, probs = seq(0.05, 0.95, 0.90))
	return ( q[1] / q[2] )
}

# find an appropriate basis for a set of N signals (data[1,],..,data[N,])
# filt = the filter type. Supported types include:
#	EXTREMAL PHASE (daublet): "haar", "d2", "d4", "d6", "d8",
#							"d10", "d12", "d14", "d16", "d18", "d20"
#	LEAST ASYMMETRIC (symmlet): "s2","s4", "s6", "s8", "s10",
#							"s12", "s14", "s16", "s18", "s20"
#	BEST LOCALIZED: "l2","l4", "l6", "l14", "l18", "l20"
#	COIFLET: "c6", "c12", "c18", "c24", "c30"
bestWavelets = function(dataForRef, lvl, filt, costType, nbCoefs, verbose=TRUE) {

	# compute wavelet packets coefficients for the "best representative" signal
	# i.e. : the median signal, for L2 norm.
	N = nrow(dataForRef) ; nd = rep(0,N)
	for (i in 1:N) nd[i] = sum(dataForRef[i,]^2)
	R = rank(nd) ; seqVect = seq(to=N)
	med = dataForRef[seqVect[R==N%/%2],]
	W <- wavDWPT(med, n.level=lvl, wavelet = filt)

	# create the cost vector
	C <- unlist(lapply(W$data, costType))
	
	# calculate the best basis (inside the dictionary)
	z <- wavBestBasis(C)
	
	# print the crystals of the best basis (optional)
	if (verbose) print(paste("W(", z$level, ",", z$osc, ")", sep=""))

	# return the corresponding basis
	bon = crystalsToBasis(W,z$level,z$osc)
	# orthonormalize the basis :
	for (i in 1:nrow(bon))
		bon[i,] = bon[i,] / sqrt(sum(bon[i,]^2))
	
	return (basisMostVar(bon,dataForRef,nbCoefs,verbose))
}

#just find the "best basis"
bestBasis = function(dataForRef, lvl, costType, nclusts, nbCoefs, dataToDistor=dataForRef, verbose=TRUE) {
	filters = c("haar", "d2", "d4", "d6", "d8", "d10", 
				"d12", "d14", "d16", "d18", "d20",
				"s2","s4", "s6", "s8", "s10",
				"s12", "s14", "s16", "s18", "s20",
				"l2","l4", "l6", "l14", "l18", "l20",
				"c6", "c12", "c18", "c24", "c30")
	bestDistor = -1.0
	bestFilt = ""

	for (filt in filters) {
		if (verbose) print(paste("current filter: ",filt))
		basis = bestWavelets(dataForRef,lvl,filt,costType,nbCoefs,FALSE)

		F = dataForRef %*% t(basis)
		k = kmeans(F,nclusts,nstart=100,iter.max=500)
		
		#compute distortion
		centers = k$centers %*% basis
		distor = distortion(dataToDistor,centers)

		if (bestDistor < 0.0 || distor < bestDistor) {
			bestDistor = distor
			bestFilt = filt
		}
	}

	if (verbose) print(paste("    best filter selected: ",bestFilt))
	basis = bestWavelets(dataForRef, lvl, bestFilt, costType, nbCoefs, FALSE)
	return (list('base'=basis,'bfilt'=bestFilt))
}


####################################################
## MAIN FUNCTION FOR THIS FILE :
####################################################

# distorsion apres k-means sur nbCoefs coefficients, SVD ou base utilisateur (ondelettes)
# eventuellement avec bestBasis pour un certain critere (2 derniers arguments)
# nwav : normal wavelets, for a given filter filt;
# bwav : best wavelet basis (try all filters, minimise entropy or..)
# fpca : K-L basis ; four : Fourier basis
basisDistor = function(data,nclusts,nbCoefs,
					basisType="nwav",basis=matrix(0,nrow=nbCoefs,ncol=ncol(data)),filt="s8",
					costType="entropy",dataForRef=data,dataToDistor=data,
					plotting=TRUE,verbose=TRUE) {
	D = ncol(data)
	lvl = floor(log(D) / log(2))
	bestFilt = ""

	if (basisType!="user" && verbose) print("basis construction...")
	# find the basis :
	if (basisType=="four") {
		basis = genFourier(dataForRef,nbCoefs,withVar=TRUE)
	} else
	if (basisType=="fpca") {
		s = svd(dataForRef)
		basis = t((s$v)[,1:min(nbCoefs,ncol(s$v))])
	} else
	if (basisType=="nwav") {
		basis = threshWav(dataForRef, lvl, filt, nbCoefs,withVar=TRUE)
	} else 
	if (basisType=="bwav") {
		bb = bestBasis(dataForRef, lvl, costType, nclusts, nbCoefs, dataToDistor,verbose)
		basis = bb$base
		bestFilt = bb$bfilt
	}

	if (verbose) print("clustering coefficients...")
	# run k-means
	cfs = data %*% t(basis)
	k = kmeans(cfs,nclusts,nstart=100,iter.max=500)

	# compute distortion (with optional plotting)
	centers = k$centers %*% basis
	distor = distortion(dataToDistor,centers)
	if (verbose) print(paste("result: average distortion = ",distor))
	if (plotting) {
		allDC = rbind(data,centers)
		rg = c(min(allDC),max(allDC))
		plotThis(data,cl=k$cluster,range=rg)
		plotThis(centers,cl=rep(1,nclusts),range=rg,lty=2,lwd=3)
	}
	if (verbose && nrow(basis)<nbCoefs) {
		print(paste("    warning: only ",nrow(basis)," coefficients used"))
	}
	return (list('bfilt'=bestFilt,'base'=basis,'disto'=distor,'ctrs'=centers,'clusts'=k$cluster))
}
