require(rgl)

# trace des evolutions de W en fonction de d : (et autres fcts utiles)

getCribs = function(ldims) {
	tops = 4*ldims
	crib = list( seq(from=1,to=tops-3,by=4),
				seq(from=2,to=tops-2,by=4),
				seq(from=3,to=tops-1,by=4),
				seq(from=4,to=tops,by=4) )
	return (crib)
}

getNumBasis = function(btype) {
	if (btype=="four") return (1)
	if (btype=="fpca") return (2)
	if (btype=="nwav") return (3)
	if (btype=="bwav") return (4)
	return (0)
}

#recupere un vecteur d'evolutions a n ou d fixes : (N : number of different values of n)
getEvol = function(distors,ldata,ldims,fixed="n",val=1,btype="four",startInd=1) {
	addCoef = getNumBasis(btype)
	crib = getCribs(ldims)
	if (fixed=="n") return ( ((distors[(4*ldims*(val-1)+1):(4*ldims*val)])[ crib[[addCoef]] ])[startInd:ldims] )
	if (fixed=="d") {
		cribD = seq(from=4*(val-1)+addCoef,to=4*(val-1)+addCoef+4*ldims*(ldata-1),by=4*ldims)
		return ( (distors[cribD])[startInd:ldata] )
	}
	return (0)
}

#recupere une matrice d'evolution de distor en fonction de n ou d (prete pr le trace) :
getEvolMatrix = function(distors,ldata,ldims,lines="n",btype="four") {
	res = matrix()
	if (lines=="n") {
		res = matrix(nrow=ldata,ncol=ldims)
		for (i in 1:ldata) res[i,] = getEvol(distors,ldata,ldims,"n",i,btype,1)
	}
	else if (lines=="d") {
		res = matrix(nrow=ldims,ncol=ldata)
		for (i in 1:ldims) res[i,] = getEvol(distors,ldata,ldims,"d",i,btype,1)
	}
	return (res)
}

#recupere un clustering pour le run k, n et d + base donnes :
getCluster = function(clusters,ldims,k,n,d,btype="four") {
	clustersM = matrix()
	if (is.list(clusters)) clustersM = clusters[[k]][[n]]
	else clustersM = clusters
	addCoef = getNumBasis(btype)
	return (clustersM[4*(d-1)+addCoef,])
}

#recupere toutes les partitions a n ou d + k, base fixe :
getClusterMatrix = function(clusters,ldata,ldims,k,fixed="n",val,btype="four") {
	res = matrix()
	ncolData = 0
	if (is.list(clusters)) {
		if (fixed=="n") ncolData = ncol(clusters[[1]][[val]])
		else ncolData = ncol(clusters[[1]][[1]])
	}
	else ncolData = ncol(clusters)
	if (fixed=="d") {
		res = matrix(nrow=ldata,ncol=ncolData)
		for (i in 1:ldata) res[i,] = getCluster(clusters,ldims,k,i,val,btype)
	}
	else if (fixed=="n") {
		res = matrix(nrow=ldims,ncol=ncolData)
		for (i in 1:ldims) res[i,] = getCluster(clusters,ldims,k,val,i,btype)
	}
	return (res)
}

#preparateur visuel pr l'interet du trace d'une nappe distor vs d ou n :
drawCurves = function(distors,ldata,ldims,valn=NULL,vald=NULL,fixed="n",val=1,startInd=1) {
	distorsM = matrix()
	if (is.list(distors)) distorsM = distors[[numN]]
	else distorsM = distors
	
	#preparations :
	evol1 = getEvol(distorsM,ldata,ldims,fixed,val,btype="four",1)
	evol2 = getEvol(distorsM,ldata,ldims,fixed,val,btype="fpca",1)
	evol3 = getEvol(distorsM,ldata,ldims,fixed,val,btype="nwav",1)
	evol4 = getEvol(distorsM,ldata,ldims,fixed,val,btype="bwav",1)
	sdraw = length(evol1)
	bounds = c(min(evol1,evol2,evol3,evol4),max(evol1,evol2,evol3,evol4))
	if (fixed=="d" && is.null(valn)) {
		valn = seq(from=100,to=3100,by=200)
#		for (i in 1:16) {
#			if (valn[i]%%6 > 0) valn[i] = valn[i] + 6 - valn[i]%%6 #correction inutile..
#		}
	}
	if (fixed=="n" && is.null(vald)) vald = seq(from=2,to=50,by=4)
	tdraw = c()
	if (fixed=="n") {
		tdraw = vald
		xlabDraw = "dimension"
	}
	else {
		tdraw = valn
		xlabDraw = "data size"
	}

	#trace distor en fonction de d ou n, pour les 4 bases :
	plot(tdraw[startInd:sdraw],evol1[startInd:sdraw],ylim=bounds,ylab="distortion",xlab=xlabDraw,cex.lab=3,lwd=2,type="l",col=1,axes=F); par(new=TRUE)
	plot(tdraw[startInd:sdraw],evol2[startInd:sdraw],ylim=bounds,ylab="distortion",xlab=xlabDraw,cex.lab=3,lwd=2,type="l",col=2,axes=F); par(new=TRUE)
	plot(tdraw[startInd:sdraw],evol3[startInd:sdraw],ylim=bounds,ylab="distortion",xlab=xlabDraw,cex.lab=3,lwd=2,type="l",col=3,axes=F); par(new=TRUE)
	plot(tdraw[startInd:sdraw],evol4[startInd:sdraw],ylim=bounds,ylab="distortion",xlab=xlabDraw,cex.lab=3,lwd=2,type="l",col=4,axes=F); par(new=TRUE)
	axis(1,at=tdraw[startInd:sdraw],labels=F)
	axis(1,at=tdraw[startInd:sdraw],col="black",lwd=2,cex.axis=3)
	axis(2,labels=F)
	axis(2,col="black",lwd=2,cex.axis=3)
}

#trace d'une nappe n, d, distor :
drawNappe = function(distors,ldata,ldims,valn=NULL,vald=NULL,btype="four",startInd=1) {
	matxTrace = matrix(nrow=0,ncol=ldims)
	if (is.character(distors)) maxTrace = as.matrix(read.table(distors))
	else {
		for (i in 1:ldata)
			matxTrace=rbind(matxTrace,getEvol(distors,ldata,ldims,fixed="n",val=i,btype,1))
	}

	if (is.null(valn))
		valn = seq(from=100,to=3100,by=200)
	if (is.null(vald))
		vald = seq(from=2,to=50,by=4)

	persp3d(x=valn,y=vald,z=matxTrace[startInd:ldata,],color=rainbow(256),xlab="data size",ylab="dimension",zlab="distortion")
}

#recuperation des matrices des centres (pour base btype, run k) :
getCentersMatrix = function(centers,ldata,ldims,nclusts,k,btype="four") {
	centersM = matrix()
	if (is.list(centers)) centersM = centers[[k]]
	else centersM = centers
	addCoef = getNumBasis(btype)
	ctrs = matrix(nrow=0,ncol=ncol(centersM))
	for (i in 1:ldata) {
		for (j in 1:ldims)
			ctrs = rbind(ctrs,centersM[((i-1)*4*nclusts*ldims+(j-1)*4*nclusts+(addCoef-1)*nclusts+1):((i-1)*4*nclusts*ldims+(j-1)*4*nclusts+(addCoef-1)*nclusts+nclusts),])
	}
	return (ctrs)
}

################################################################################
## script de sauvegarde apres N runs :

#ATTENTION : ne fonctionne qu'apres execution de "runTests.R"
saveEmAll = function(ldata=16,ldims=13,valn=NULL,vald=NULL,N=1) {

	#1) les distorsions
	#CCT :
	ev1 = getEvolMatrix(distorsCCT,ldata,ldims,lines="n",btype="four")
	ev2 = getEvolMatrix(distorsCCT,ldata,ldims,lines="n",btype="fpca")
	ev3 = getEvolMatrix(distorsCCT,ldata,ldims,lines="n",btype="nwav")
	ev4 = getEvolMatrix(distorsCCT,ldata,ldims,lines="n",btype="bwav")
	write(t(ev1),"distortions/CCTdistor_four.txt",ncolumns=ncol(ev1))
	write(t(ev2),"distortions/CCTdistor_fpca.txt",ncolumns=ncol(ev2))
	write(t(ev3),"distortions/CCTdistor_nwav.txt",ncolumns=ncol(ev3))
	write(t(ev4),"distortions/CCTdistor_bwav.txt",ncolumns=ncol(ev4))
	#cat :
	ev1 = getEvolMatrix(cbc$distors,ldata=1,ldims,lines="n",btype="four")
	ev2 = getEvolMatrix(cbc$distors,ldata=1,ldims,lines="n",btype="fpca")
	ev3 = getEvolMatrix(cbc$distors,ldata=1,ldims,lines="n",btype="nwav")
	ev4 = getEvolMatrix(cbc$distors,ldata=1,ldims,lines="n",btype="bwav")
	write(t(rbind(ev1,ev2,ev3,ev4)),"distortions/Cat_distors.txt",ncolumns=ncol(ev1))
	#bem1 :
	ev1 = getEvolMatrix(cbb1$distors,ldata=1,ldims,lines="n",btype="four")
	ev2 = getEvolMatrix(cbb1$distors,ldata=1,ldims,lines="n",btype="fpca")
	ev3 = getEvolMatrix(cbb1$distors,ldata=1,ldims,lines="n",btype="nwav")
	ev4 = getEvolMatrix(cbb1$distors,ldata=1,ldims,lines="n",btype="bwav")
	write(t(rbind(ev1,ev2,ev3,ev4)),"distortions/Bem1_distors.txt",ncolumns=ncol(ev1))
	#bem2 :
	ev1 = getEvolMatrix(cbb2$distors,ldata=1,ldims,lines="n",btype="four")
	ev2 = getEvolMatrix(cbb2$distors,ldata=1,ldims,lines="n",btype="fpca")
	ev3 = getEvolMatrix(cbb2$distors,ldata=1,ldims,lines="n",btype="nwav")
	ev4 = getEvolMatrix(cbb2$distors,ldata=1,ldims,lines="n",btype="bwav")
	write(t(rbind(ev1,ev2,ev3,ev4)),"distortions/Bem1_distors.txt",ncolumns=ncol(ev1))
	
	if (is.null(valn))
		valn = seq(from=100,to=3100,by=200)
	if (is.null(vald))
		vald = seq(from=2,to=50,by=4)

	#2) les clusters (pour un run aleatoire, sans grandes consequences)
	numRun = sample(N,1)
	#CCT :
	for (i in 1:ldata) {
		cm1 = getClusterMatrix(clusters_cct,ldata,ldims,k=numRun,fixed="n",val=i,btype="four")
		cm2 = getClusterMatrix(clusters_cct,ldata,ldims,k=numRun,fixed="n",val=i,btype="fpca")
		cm3 = getClusterMatrix(clusters_cct,ldata,ldims,k=numRun,fixed="n",val=i,btype="nwav")
		cm4 = getClusterMatrix(clusters_cct,ldata,ldims,k=numRun,fixed="n",val=i,btype="bwav")
		name = paste("clusters/CCTclusters",valn[i],".txt",sep="")
		write(t(rbind(cm1,cm2,cm3,cm4)),name,ncolumns=ncol(cm1))
	}
	#cat :
	cm1 = getClusterMatrix(cbc$clusters,ldata=1,ldims,k=1,fixed="n",val=1,btype="four")
	cm2 = getClusterMatrix(cbc$clusters,ldata=1,ldims,k=1,fixed="n",val=1,btype="fpca")
	cm3 = getClusterMatrix(cbc$clusters,ldata=1,ldims,k=1,fixed="n",val=1,btype="nwav")
	cm4 = getClusterMatrix(cbc$clusters,ldata=1,ldims,k=1,fixed="n",val=1,btype="bwav")
	write(t(rbind(cm1,cm2,cm3,cm4)),"clusters/CatClusters.txt",ncolumns=ncol(cm1))
	#bem1 :
	cm1 = getClusterMatrix(cbb1$clusters,ldata=1,ldims,k=1,fixed="n",val=1,btype="four")
	cm2 = getClusterMatrix(cbb1$clusters,ldata=1,ldims,k=1,fixed="n",val=1,btype="fpca")
	cm3 = getClusterMatrix(cbb1$clusters,ldata=1,ldims,k=1,fixed="n",val=1,btype="nwav")
	cm4 = getClusterMatrix(cbb1$clusters,ldata=1,ldims,k=1,fixed="n",val=1,btype="bwav")
	write(t(rbind(cm1,cm2,cm3,cm4)),"clusters/Bem1Clusters.txt",ncolumns=ncol(cm1))
	#bem2 :
	cm1 = getClusterMatrix(cbb2$clusters,ldata=1,ldims,k=1,fixed="n",val=1,btype="four")
	cm2 = getClusterMatrix(cbb2$clusters,ldata=1,ldims,k=1,fixed="n",val=1,btype="fpca")
	cm3 = getClusterMatrix(cbb2$clusters,ldata=1,ldims,k=1,fixed="n",val=1,btype="nwav")
	cm4 = getClusterMatrix(cbb2$clusters,ldata=1,ldims,k=1,fixed="n",val=1,btype="bwav")
	write(t(rbind(cm1,cm2,cm3,cm4)),"clusters/Bem2Clusters.txt",ncolumns=ncol(cm1))

	#3) les centres (pour le meme run aleatoire)
	#CCT :
	cm1 = getCentersMatrix(centers_cct,ldata,ldims,nclusts=6,k=numRun,btype="four")
	cm2 = getCentersMatrix(centers_cct,ldata,ldims,nclusts=6,k=numRun,btype="fpca")
	cm3 = getCentersMatrix(centers_cct,ldata,ldims,nclusts=6,k=numRun,btype="nwav")
	cm4 = getCentersMatrix(centers_cct,ldata,ldims,nclusts=6,k=numRun,btype="bwav")
	write(t(cm1),"centers/CCTcenters_four.txt",ncolumns=ncol(cm1))
	write(t(cm2),"centers/CCTcenters_fpca.txt",ncolumns=ncol(cm2))
	write(t(cm3),"centers/CCTcenters_nwav.txt",ncolumns=ncol(cm3))
	write(t(cm4),"centers/CCTcenters_bwav.txt",ncolumns=ncol(cm4))
	#cat :
	cm1 = getCentersMatrix(cbc$centers,ldata=1,ldims,nclusts=2,k=1,btype="four")
	cm2 = getCentersMatrix(cbc$centers,ldata=1,ldims,nclusts=2,k=1,btype="fpca")
	cm3 = getCentersMatrix(cbc$centers,ldata=1,ldims,nclusts=2,k=1,btype="nwav")
	cm4 = getCentersMatrix(cbc$centers,ldata=1,ldims,nclusts=2,k=1,btype="bwav")
	write(t(rbind(cm1,cm2,cm3,cm4)),"centers/CatCenters.txt",ncolumns=ncol(cm1))
	#bem1 :
	cm1 = getCentersMatrix(cbb1$centers,ldata=1,ldims,nclusts=2,k=1,btype="four")
	cm2 = getCentersMatrix(cbb1$centers,ldata=1,ldims,nclusts=2,k=1,btype="fpca")
	cm3 = getCentersMatrix(cbb1$centers,ldata=1,ldims,nclusts=2,k=1,btype="nwav")
	cm4 = getCentersMatrix(cbb1$centers,ldata=1,ldims,nclusts=2,k=1,btype="bwav")
	write(t(rbind(cm1,cm2,cm3,cm4)),"centers/Bem1Centers.txt",ncolumns=ncol(cm1))
	#bem2 :
	cm1 = getCentersMatrix(cbb2$centers,ldata=1,ldims,nclusts=3,k=1,btype="four")
	cm2 = getCentersMatrix(cbb2$centers,ldata=1,ldims,nclusts=3,k=1,btype="fpca")
	cm3 = getCentersMatrix(cbb2$centers,ldata=1,ldims,nclusts=3,k=1,btype="nwav")
	cm4 = getCentersMatrix(cbb2$centers,ldata=1,ldims,nclusts=3,k=1,btype="bwav")
	write(t(rbind(cm1,cm2,cm3,cm4)),"centers/Bem2Centers.txt",ncolumns=ncol(cm1))
}
